package ru.ifmo.qbeing.global;

import universum.bi.*;
import ru.ifmo.qbeing.common.LocationInfo;
import ru.ifmo.qbeing.common.BeingUtil;
import ru.ifmo.qbeing.common.BehaviourConstants;
import ru.ifmo.qbeing.QBeing;

import java.util.*;

public class UniverseInfo {
    protected Map<Location, LocationInfo> knownLocations;
    protected Map<Location, LocationInfo> foodLocations;
    protected Map<QBeing, Location> beingLocations;
    protected Map<QBeing, Location> exploreCenters;
    protected Map<Location, Set<QBeing>> beingsByLocation;
    protected Map<Location, Set<QBeing>> beingFollowingLocation;
    protected Map<Location, Set<QBeing>> beingSupportLocation;

    protected static UniverseInfo instance = null;

    protected UniverseInfo() {
        knownLocations = new HashMap<Location, LocationInfo>();
        foodLocations = new HashMap<Location, LocationInfo>();
        beingLocations = new HashMap<QBeing, Location>();
        exploreCenters = new HashMap<QBeing, Location>();
        beingsByLocation = new HashMap<Location, Set<QBeing>>();
        beingFollowingLocation = new HashMap<Location, Set<QBeing>>();
        beingSupportLocation = new HashMap<Location, Set<QBeing>>();
    }

    public static void initInstance() {
        instance = new UniverseInfo();
    }

    public static UniverseInfo instance() {
        if (instance == null) {
            initInstance();
        }
        return instance;
    }

    public int getExploredLocationCount() {
        return knownLocations.keySet().size();
    }

    public synchronized void updateKnownLocations(QBeing me, BeingInterface bi) {
        List<PointInfo> neighbourInfo = bi.getNeighbourInfo(me);
        updateKnownLocation(me, bi.getPointInfo(me));
        for (PointInfo pi : neighbourInfo) {
            updateKnownLocation(me, pi);
        }
    }

    public synchronized void updateKnownLocation(QBeing me, PointInfo pi) {
        knownLocations.put(pi.getLocation(), new LocationInfo(pi, me));
        if ((pi.getMaxCount(me) > 0 && pi.getGrowthRate(me) > 0.001) || pi.getMaxCount(me) == -1) {
            foodLocations.put(pi.getLocation(), new LocationInfo(pi, me));
        }
    }

    public synchronized void updateBeingLocation(QBeing me, BeingInterface bi, Location newLocation) {
        Location oldLoc = beingLocations.get(me);
        removeBeingFromLocation(me, oldLoc);

        if (newLocation != null) {
            addBeingToLocationMap(beingsByLocation, me, newLocation);
            beingLocations.put(me, newLocation);
        }
    }

    public synchronized void setExplorerCenter(QBeing me, Location location) {
        exploreCenters.put(me, location);
    }

    public synchronized Location getExploreCenters(QBeing me) {
        return exploreCenters.get(me);
    }

    public synchronized Collection<Location> filterVisited(BeingInterface bi, List<Location> reachable) {
        HashSet<Location> loc = new HashSet<Location>();
        for (Location l : reachable) {
            if (!knownLocations.containsKey(l) && !findVisibleLocation(bi, l)) {
                loc.add(l);
            }
        }
        return loc;
    }

    private synchronized boolean findVisibleLocation(BeingInterface bi, Location l) {
        for (Location loc : beingLocations.values()) {
            if (bi.distance(loc, l) <= 1) {
                return true;
            }
        }
        return false;
    }

    public synchronized void removeBeingFromLocation(QBeing me, Location oldLoc) {
        beingLocations.remove(me);
        removeBeingFromLocationMap(beingsByLocation, me, oldLoc);
    }

    private void addBeingToLocationMap(Map<Location, Set<QBeing>> locMap, QBeing me, Location newLocation) {
        Set<QBeing> beings = locMap.get(newLocation);
        if (beings == null) {
            beings = new HashSet<QBeing>();
        }
        beings.add(me);
        locMap.put(newLocation, beings);
    }

    private void removeBeingFromLocationMap(Map<Location, Set<QBeing>> locMap, QBeing me, Location oldLoc) {
        if (oldLoc != null) {
            Set<QBeing> beings = locMap.get(oldLoc);
            if (beings != null) {
                beings.remove(me);
            }
        }
    }

    public synchronized float getAvailEnergy(QBeing me, LocationInfo pi) {
        Location loc = pi.location;
        float count = pi.count;
        float energyAlreadyConsumed = getConsumedAt(me, loc);
        return (count - energyAlreadyConsumed) < 0 ? 0 : count - energyAlreadyConsumed;
    }

    private float getConsumedAt(QBeing me, Location loc) {
        float consumed = 0;
        Set<QBeing> beings = beingsByLocation.get(loc);
        if (beings != null) {
            for (QBeing being : beings) {
                if (being != me)
                    consumed += being.beingWillEat();
            }
        }
        Set<QBeing> beingsArriving = beingFollowingLocation.get(loc);
        if (beingsArriving != null) {
            for (QBeing being : beingsArriving) {
                if (being != me)
                    consumed += being.beingWillEat();
            }
        }
        return consumed;
    }

    public synchronized boolean isEnoughEnergy(QBeing me, LocationInfo pi, float energyEnough) {
        return getAvailEnergy(me, pi) >= energyEnough;
    }

    public synchronized Location findEnoughEnergy(QBeing me, List<PointInfo> neighbourInfo, float energyEnough) {
        boolean first = true;
        float max = 0;
        Location bestLoc = null;
        for (PointInfo pi : neighbourInfo) {
            float count = getAvailEnergy(me, new LocationInfo(pi, me));
            if (count >= energyEnough) {
                if (count > max || first) {
                    max = count;
                    bestLoc = pi.getLocation();
                }
                first = false;
            }
        }
        return bestLoc;
    }

    public synchronized Location findReachableEnergyPoint(QBeing me, BeingInterface bi, float energyAmount) {
        return findReachableEnergyPoint(me, bi, energyAmount, false);
    }

    public synchronized Location findReachableEnergyPoint(QBeing me, BeingInterface bi, float energyAmount, boolean greedy) {
        Location myLocation = bi.getLocation(me);
        float energyForMoving = me.getEnergy() - Constants.K_emin * me.getMass();
        int myMaxSteps = (int) Math.floor(energyForMoving / (Constants.K_movecost * me.getSpeed()));

        float maxEnergy = 0;
        float stepsForBest = 0;
        Location bestLocation = null;
        boolean first = true;
        for (Map.Entry<Location, LocationInfo> entry : foodLocations.entrySet()) {
            float energyAvail = greedy ? entry.getValue().count : getAvailEnergy(me, entry.getValue());
            if (energyAvail >= energyAmount * me.getMass()) {
                // - ftr: improve (add move cost)
                int steps = (int) (Math.ceil((float) BeingUtil.getStepsCount(bi, myLocation, entry.getKey())));
                if (steps <= myMaxSteps && steps * Constants.K_movecost < energyAvail) {
                    if (maxEnergy < energyAvail ||
                            (Math.abs(maxEnergy - energyAvail) <= BehaviourConstants.EQUAL_FEED_ENERGY
                                    && steps < stepsForBest) ||
                            first) {
                        stepsForBest = steps;
                        maxEnergy = energyAvail;
                        bestLocation = entry.getKey();
                    }
                    first = false;
                }
            }
        }
        return bestLocation;
    }

    public synchronized void updateEnergyAtKnownLocations() {
        for (LocationInfo li : knownLocations.values()) {
            li.count += li.growthRate;
        }
    }

    public synchronized void setBeingFollowingLocation(QBeing me, Location loc) {
        addBeingToLocationMap(beingFollowingLocation, me, loc);
    }

    public synchronized void removeBeingFollowingLocation(QBeing me, Location loc) {
        removeBeingFromLocationMap(beingFollowingLocation, me, loc);
    }

    public synchronized float getAvailEnergyAtLoc(QBeing me, Location loc) {
        LocationInfo li = knownLocations.get(loc);
        return getAvailEnergy(me, li);
    }

    public float getSupport(Location location) {
        float force = 0;
        Set<QBeing> beings = beingSupportLocation.get(location);
        if (beings != null) {
            for (QBeing being : beings) {
                force += being.getForce();
            }
        }
        return force;
    }

    public void setSupportBeing(QBeing b, Location location) {
        addBeingToLocationMap(beingSupportLocation, b, location);
    }

    public void removeSupportBeing(QBeing b, Location location) {
        removeBeingFromLocationMap(beingSupportLocation, b, location);
    }
}
